#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

import numpy as np

from cdflib import CDF, cdfepoch
from dateutil import parser as date_parser

from ..pyrf import ts_skymap, datetime_to_tt2000


def get_dist(file_path, cdf_name, tint):
    """Read field named cdf_name in file and convert to velocity distribution function.

    Parameters
    ----------
    file_path : str
        Path of the cdf file.

    cdf_name : str
        Name of the target variable in the cdf file.

    tint : list of str
        Time interval.

    Returns
    -------
    out : xarray.Dataset
        Time series of the velocity distribution function if the target specie in the selected time
        interval.

    """

    tmmode = cdf_name.split("_")[-1]

    tint = list(map(date_parser.parse, tint))
    tint = list(map(datetime_to_tt2000, tint))
    tint = list(map(cdfepoch.parse, tint))

    with CDF(file_path) as f:
        if tmmode == "brst":
            depend0_key = f.varattsget(cdf_name)["DEPEND_0"]
            depend1_key = f.varattsget(cdf_name)["DEPEND_1"]
            depend2_key = f.varattsget(cdf_name)["DEPEND_2"]
            depend3_key = f.varattsget(cdf_name)["DEPEND_3"]

            t = f.varget(depend0_key, starttime=tint[0], endtime=tint[1])
            t = cdfepoch.to_datetime(t, to_np=True)

            if not t.size:
                return None

            dist = f.varget(cdf_name, starttime=tint[0], endtime=tint[1])
            dist = np.transpose(dist, [0, 3, 1, 2])
            ph = f.varget(depend1_key, starttime=tint[0], endtime=tint[1])
            th = f.varget(depend2_key)
            en = f.varget(depend3_key, starttime=tint[0], endtime=tint[1])

            en0_name = "_".join([cdf_name.split("_")[0], cdf_name.split("_")[1], "energy0",
                                 cdf_name.split("_")[-1]])
            en1_name = "_".join([cdf_name.split("_")[0], cdf_name.split("_")[1], "energy1",
                                 cdf_name.split("_")[-1]])
            d_en_name = "_".join([cdf_name.split("_")[0], cdf_name.split("_")[1], "energy_delta",
                                  cdf_name.split("_")[-1]])
            e_step_table_name = "_".join([cdf_name.split("_")[0], cdf_name.split("_")[1],
                                          "steptable_parity", cdf_name.split("_")[-1]])

            step_table = f.varget(e_step_table_name, starttime=tint[0], endtime=tint[1])
            if d_en_name in f.cdf_info()["zVariables"]:
                delta_plus_var = f.varget(d_en_name, starttime=tint[0], endtime=tint[1])
                delta_minus_var = f.varget(d_en_name, starttime=tint[0], endtime=tint[1])

            if en0_name not in f.cdf_info()["zVariables"]:
                energy0 = en[1, :]
                energy1 = en[0, :]
            else:
                energy0 = f.varget(en0_name)
                energy1 = f.varget(en1_name)

            res = ts_skymap(t, dist, None, ph, th, energy0=energy0, energy1=energy1,
                            esteptable=step_table)

            if "delta_plus_var" in locals() and "delta_minus_var" in locals():
                res.attrs["delta_energy_minus"] = delta_minus_var
                res.attrs["delta_energy_plus"] = delta_plus_var

            res.attrs = {**res.attrs, **f.varattsget(cdf_name)}

            for k in f.cdf_info():
                res.attrs[k] = f.cdf_info()[k]

            res.attrs["tmmode"] = tmmode
            if "_dis_" in cdf_name:
                res.attrs["species"] = "ions"
            else:
                res.attrs["species"] = "electrons"

        elif tmmode == "fast":
            depend0_key = f.varattsget(cdf_name)["DEPEND_0"]
            depend1_key = f.varattsget(cdf_name)["DEPEND_1"]
            depend2_key = f.varattsget(cdf_name)["DEPEND_2"]
            depend3_key = f.varattsget(cdf_name)["DEPEND_3"]

            t = f.varget(depend0_key, starttime=tint[0], endtime=tint[1])

            dist = f.varget(cdf_name, starttime=tint[0], endtime=tint[1])
            dist = np.transpose(dist, [0, 3, 1, 2])
            ph = f.varget(depend1_key)
            th = f.varget(depend2_key)
            en = f.varget(depend3_key, starttime=tint[0], endtime=tint[1])
            res = ts_skymap(t, dist, en, ph, th)

            for k in f.varattsget(cdf_name):
                res.attrs[k] = f.varattsget(cdf_name)[k]

            for k in f.cdf_info():
                res.attrs[k] = f.cdf_info()[k]

            res.attrs["tmmode"] = tmmode
            if "_dis_" in cdf_name:
                res.attrs["species"] = "ions"
            else:
                res.attrs["species"] = "electrons"
    return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""norm.py
@author: Louis Richard
"""

import numpy as np


def norm(inp):
    """Computes the magnitude of the input field.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input field.

    Returns
    -------
    out : xarray.DataArray
        Time series of the magnitude of the input field.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft index

    >>> mms_id = 1

    Load magnetic field

    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)

    Compute magnitude of the magnetic field

    >>> b_mag = pyrf.norm(b_xyz)

    """

    out = np.sqrt(np.sum(inp ** 2, axis=1))

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

import os
import warnings
import numpy as np
import pandas as pd

from scipy import interpolate
from astropy.time import Time


def igrf(time, flag):
    """Returns magnetic dipole latitude and longitude of the IGRF model

    Parameters
    ----------
    time : ndarray
        Times in unix format.

    flag : str
        Default is dipole.


    Returns
    -------
    lambda : ndarray
        latitude

    phi : ndarray
        longitude

    """

    # Root path
    # root_path = os.getcwd()
    root_path = os.path.dirname(os.path.abspath(__file__))

    # File path
    file_name = "data/igrf13coeffs.csv"

    # file reading
    df = pd.read_csv(os.sep.join([root_path, file_name]))

    # construct IGRF coefficient matrices
    a = df.loc[0]
    years_igrf = a[3:].to_list()
    years_igrf[-1] = float(years_igrf[-1].split("-")[0]) + 5.

    # read in all IGRF coefficients from file
    i_igrf = df.loc[1:, ].values
    i_igrf[:, -1] = i_igrf[:, -2] + 5. * i_igrf[:, -1].astype(float)
    h_igrf = i_igrf[i_igrf[:, 0] == "h", 1:].astype(float)
    g_igrf = i_igrf[i_igrf[:, 0] == "g", 1:].astype(float)

    # timeVec = irf_time(t,'vector');
    # yearRef = timeVec(:,1);
    year_ref = Time(time, format="unix").datetime64.astype("datetime64[Y]")
    year_ref = year_ref.astype(int) + 1970
    year_ref_isot = list(map(lambda x: f"{x}-01-01T00:00:00", year_ref))
    year_ref_unix = Time(year_ref_isot, format="isot").unix

    if np.min(year_ref) < np.min(years_igrf):
        message = "requested time is earlier than the first available IGRF " \
                  "model from extrapolating in past.. "
        warnings.warn(message, category=UserWarning)

    year = year_ref + (time - year_ref_unix) / (365.25 * 86400)

    assert flag == "dipole", "input flag is not recognized"

    tck_g0_igrf = interpolate.interp1d(years_igrf, g_igrf[0, 2:],
                                       kind="linear", fill_value="extrapolate")
    tck_g1_igrf = interpolate.interp1d(years_igrf, g_igrf[1, 2:],
                                       kind="linear", fill_value="extrapolate")
    tck_h0_igrf = interpolate.interp1d(years_igrf, h_igrf[0, 2:],
                                       kind="linear", fill_value="extrapolate")

    g01 = tck_g0_igrf(year)
    g11 = tck_g1_igrf(year)
    h11 = tck_h0_igrf(year)
    lambda_ = np.arctan(h11 / g11)
    phi = np.pi / 2
    phi -= np.arcsin((g11 * np.cos(lambda_) + h11 * np.sin(lambda_)) / g01)

    return np.rad2deg(lambda_), np.rad2deg(phi)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""eb_nrf.py
@author: Louis Richard
"""

import numpy as np
import xarray as xr

from .resample import resample
from .dot import dot
from .normalize import normalize
from .cross import cross


def eb_nrf(e_xyz, b_xyz, v_xyz, flag=0):
    """Find E and B in MP system given B and MP normal vector.

    Parameters
    ----------
    e_xyz : xarray.DataArray
        Time series of the electric field.

    b_xyz : xarray.DataArray
        Time series of the magnetic field.

    v_xyz : xarray.DataArray
        Normal vector.

    flag : int or ndarray
        to fill.

    Returns
    -------
    out : xarray.DataArray
        to fill.

    """

    assert isinstance(flag, (int, np.ndarray)), "Invalid flag type"

    if isinstance(flag, int):
        flag_cases = ["a", "b"]
        flag_case = flag_cases[flag]
        l_direction = None

    else:
        assert np.size(flag) == 3
        l_direction = flag
        flag_case = "c"

    if flag_case == "a":
        b_data = resample(b_xyz, e_xyz).data

        n_l = b_data / np.linalg.norm(b_data, axis=0)[:, None]
        n_n = np.cross(np.cross(b_data, v_xyz), b_data)
        n_n = n_n / np.linalg.norm(n_n)[:, None]
        n_m = np.cross(n_n, n_l)  # in (vn x b) direction

    elif flag_case == "b":
        n_n = v_xyz / np.linalg.norm(v_xyz)
        n_m = normalize(cross(n_n, b_xyz.mean(dim="time")))
        n_l = cross(n_m, n_n)

    else:
        n_n = normalize(v_xyz)
        n_m = normalize(np.cross(n_n, l_direction))
        n_l = cross(n_m, n_n)

    # estimate e in new coordinates
    e_lmn = np.hstack([dot(e_xyz, vec) for vec in [n_l, n_m, n_n]])

    out = xr.DataArray(e_xyz.time.data, e_lmn, e_xyz.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

import numpy as np
import xarray as xr

from scipy import constants

from ..pyrf import (filt, calc_dt, resample, convert_fac, ts_scalar, extend_tint, time_clip,
                    ts_vec_xyz)


def lh_wave_analysis(tints, e_xyz, b_scm, b_xyz, n_e, **kwargs):
    """
    Calculates lower-hybrid wave properties from MMS data

    Parameters
    ----------
    tints : list of str
        Time interval

    e_xyz : xarray.DataArray
        Time series pf the electric field

    b_scm : xarray.DataArray
        Time series of the fluctuations of the magnetic field

    b_xyz : xarray.DataArray
        Time series of the background magnetic field

    n_e : xarray.DataArray
        Time series of the number density

    **kwargs : dict
        Hash table of keyword arguments with :
            * lhfilt : float or int or list of float or list of int
                Filter for LH fluctuations. For one element it is the minimum frequency in the
                highpass filter. For two elements the fields are bandpassed between the frequencies.

            * blpass : float or int
                Set maximum frequency for low-pass filter of background magnetic field (FGM)

    Returns
    -------
    phi_eb : xarray.DataArray
        to fill

    v_best : ndarray
        to fill

    dir_best : ndarray
        to fill

    thetas : ndarray
        to fill

    corrs : ndarray
        to fill

    Examples
    --------
    >>> from pyrfu.mms import get_data, lh_wave_analysis

    Define time intervals

    >>> tint_long = ["2015-12-14T01:17:39.000", "2015-12-14T01:17:43.000"]
    >>> tint_zoom = ["2015-12-14T01:17:40.200","2015-12-14T01:17:41.500"]

    Load fields and density

    >>> b_gse = get_data("b_gse_fgm_brst_l2", tint_long, 2)
    >>> e_gse = get_data("e_gse_edp_brst_l2", tint_long, 2)
    >>> b_scm = get_data("b_gse_scm_brst_l2", tint_long, 2)
    >>> n_e = get_data("ne_fpi_brst_l2", tint_long, 2)

    Lower Hybrid Waves Analysis

    >>> opt = dict(lhfilt=[5, 100])
    >>> phi_eb, v_best, dir_best, theta, _ = lh_wave_analysis(tint, e_xyz, b_scm, b_xyz, n_e, **opt)

    """

    # Default band passes
    min_freq = 10
    max_freq = 0
    lowpass_b_xyz = 2

    if "lhfilt" in kwargs:
        if isinstance(kwargs["lhfilt"], (float, int)):
            min_freq = kwargs["lhfilt"]
        elif isinstance(kwargs["lhfilt"], (list, np.ndarray)) and kwargs["lhfilt"]:
            min_freq = kwargs["lhfilt"][0]
            max_freq = kwargs["lhfilt"][1]
        else:
            raise ValueError("lhfilt option not recognized")

    if "blpass" in kwargs:
        if isinstance(kwargs["blpass"], (float, int)):
            lowpass_b_xyz = kwargs["blpass"]
        else:
            raise ValueError("blpass option not recognized")

    # Bandpass filter data
    b_xyz = filt(b_xyz, 0, lowpass_b_xyz, 5)
    e_xyz = resample(e_xyz, b_scm)
    n_e = resample(n_e, b_scm)
    b_xyz = resample(b_xyz, b_scm)
    b_scm_fac = convert_fac(b_scm, b_xyz, [1, 0, 0])
    b_scm_fac = filt(b_scm_fac, min_freq, max_freq, 5)
    e_xyz = filt(e_xyz, min_freq, max_freq, 5)

    q_e = constants.elementary_charge
    mu0 = constants.mu_0

    b_mag = np.linalg.norm(b_xyz, axis=1)
    phi_b = (b_scm_fac.data[:, 2]) * b_mag * 1e-18 / (n_e.data * q_e * mu0 * 1e6)
    phi_b = ts_scalar(b_scm_fac.time.data, phi_b)

    # short buffer so phi_E does not begin at zero.
    tint = extend_tint(tints, [-.2, .2])

    e_xyz = time_clip(e_xyz, tint)
    phi_bs = time_clip(phi_b, tints)

    # Rotate e_xyz into field-aligned coordinates
    b_xyz_tc = time_clip(b_xyz, tints)
    b_mean = np.mean(b_xyz_tc.data, axis=0)
    b_vec = b_mean / np.linalg.norm(b_mean)
    r_temp = [1, 0, 0]
    bxr = np.cross(b_vec, r_temp)
    bxr /= np.linalg.norm(bxr)
    bxrxb = np.cross(bxr, b_vec)
    er1 = e_xyz.data[:, 0] * bxrxb[0] + e_xyz.data[:, 1] * bxrxb[1] + e_xyz.data[:, 2] * bxrxb[2]
    er2 = e_xyz.data[:, 0] * bxr[0] + e_xyz.data[:, 1] * bxr[1] + e_xyz.data[:, 2] * bxr[2]
    er3 = e_xyz.data[:, 0] * b_vec[0] + e_xyz.data[:, 1] * b_vec[1] + e_xyz.data[:, 2] * b_vec[2]

    e_fac = ts_vec_xyz(e_xyz.time.data, np.vstack([er1, er2, er3]).T)

    # Find best direction
    dt_e_fac = calc_dt(e_fac)
    thetas = np.linspace(0, 2 * np.pi, 361)
    corrs = np.zeros(len(thetas))

    for i, theta in enumerate(thetas):
        e_temp = np.cos(theta) * e_fac.data[:, 0] + np.sin(theta) * e_fac.data[:, 1]

        phi_temp = ts_scalar(e_xyz.time.data, np.cumsum(e_temp) * dt_e_fac)
        phi_temp = time_clip(phi_temp, tints)
        phi_temp -= np.mean(phi_temp)

        corrs[i] = np.corrcoef(phi_bs.data, phi_temp.data)

    corrpos = np.argmax(corrs)
    e_best = np.cos(thetas[corrpos]) * e_fac.data[:, 0] + np.sin(thetas[corrpos]) * e_fac.data[:, 1]
    e_best = ts_scalar(e_xyz.time.data, e_best)
    phi_best = ts_scalar(e_xyz.time.data, np.cumsum(e_best) * dt_e_fac)
    phi_best = time_clip(phi_best, tints)
    phi_best -= np.mean(phi_best)
    theta_best = thetas[corrpos]
    dir_best = bxrxb * np.cos(theta_best) + bxr * np.sin(theta_best)

    # Find best speed
    # Maximum velocity may need to be increased in rare cases
    vph_vec = np.linspace(1e1, 5e2, 491)
    corr_v = np.zeros(len(vph_vec))

    for i, vph in enumerate(vph_vec):
        phi_e_temp = phi_best.data * vph
        corr_v[i] = np.sum(np.abs(phi_e_temp - phi_bs.data) ** 2)

    corr_vpos = np.argmin(corr_v)
    phi_e_best = phi_best.data * vph_vec[corr_vpos]
    phi_e_best = ts_scalar(phi_bs.time.data, phi_e_best)
    v_best = vph_vec[corr_vpos]

    options = dict(coords=[phi_bs.time, ["Ebest", "Bs"]], dims=["time", "comp"])
    phi_eb = xr.DataArray(np.vstack([phi_e_best.data, phi_bs.data]).T, **options)

    return phi_eb, v_best, dir_best, thetas, corrs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""mean.py
@author: Louis Richard
"""

import numpy as np
import xarray as xr

from .resample import resample
from .cross import cross
from .normalize import normalize
from .ts_vec_xyz import ts_vec_xyz


def mean(inp, r_xyz, b_xyz, dipole_axis):
    """Put inp into mean field coordinates defined by position vector
    r and magnetic field b if earth magnetic dipole axis z is given
    then  uses another algorithm (good for auroral passages)

    Parameters
    ----------
    inp : xarray.DataArray
        Input field to put into MF coordinates.

    r_xyz : xarray.DataArray
        Time series of the spacecraft position.

    b_xyz : xarray.DataArray
        Time series of the background magnetic field.

    dipole_axis : xarray.DataArray
        Earth magnetic dipole axis.

    Returns
    -------
    out : xarray.DataArray
        Input field in mean field coordinates.

    """

    if dipole_axis is not None:
        assert isinstance(dipole_axis, xr.DataArray)
        flag_dipole = True

        if len(dipole_axis) != len(inp):
            dipole_axis = resample(dipole_axis, inp)

    else:
        flag_dipole = False
        dipole_axis = None

    if len(r_xyz) != len(inp):
        r_xyz = resample(r_xyz, inp)

    if len(b_xyz) != len(inp):
        b_xyz = resample(b_xyz, inp)

    b_hat = normalize(b_xyz)

    if not flag_dipole:
        bxr = cross(b_hat, r_xyz)
        bxr /= np.linalg.norm(bxr, axis=1)[:, None]
    else:
        fact = -1 * np.ones(len(b_xyz))
        fact[np.sum(b_xyz * r_xyz) > 0] = 1
        bxr = np.cross(dipole_axis, b_xyz) * fact[:, None]
        bxr /= np.linalg.norm(bxr, axis=1)[:, None]

    bxrxb = np.cross(bxr, b_hat)

    out_data = np.zeros(inp.data.shape)
    out_data[:, 0] = np.sum(bxrxb * inp, axis=1)
    out_data[:, 1] = np.sum(bxr * inp, axis=1)
    out_data[:, 2] = np.sum(b_hat * inp, axis=1)

    out = ts_vec_xyz(inp.time.data, out_data, inp.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.


import numpy as np
import xarray as xr

from cdflib import CDF, cdfepoch
from dateutil import parser as date_parser

from ..pyrf import datetime_to_tt2000


def get_epochs(file, cdf_name, tint):
    """Read time data and its attributes associated to the variable named `cdf_name` in `file`.

    Parameters
    ----------
    file : cdflib.cdfread.CDF
        cdf file

    cdf_name : str
        Name of the target variable in the cdf file

    tint : list of str
        Time interval

    Returns
    -------
    out : dict
        Hash table with the data of the time and its meta-data.

    """

    depend0_key = file.varattsget(cdf_name)["DEPEND_0"]

    out = {"data": file.varget(depend0_key, starttime=tint[0], endtime=tint[1])}

    if file.varinq(depend0_key)["Data_Type_Description"] == "CDF_TIME_TT2000":
        out["data"] = cdfepoch.to_datetime(out["data"], to_np=True)

    out["atts"] = {}

    for k in file.varattsget(depend0_key):
        out["atts"][k] = file.varattsget(depend0_key)[k]
        if isinstance(out["atts"][k], str) and out["atts"][k] in file.cdf_info()["zVariables"] \
                and not k == "LABLAXIS":
            if out["atts"][k] not in ["Epoch_MINUS", "Epoch_PLUS"]:

                try:
                    # If array
                    out["atts"][k] = file.varget(out["atts"][k], starttime=tint[0], endtime=tint[1])
                except IndexError:
                    # If scalar
                    out["atts"][k] = file.varget(out["atts"][k])

    return out


def get_depend_attributes(file, depend_key, tint):
    """Get dependy attributes

    Parameters
    ----------
    file : cdflib.cdfread.CDF
        cdf file

    depend_key : str
        Name of the variable depency in the cdf file

    tint : list of str
        Time interval

    Returns
    -------
    attributes : dict
        Hash table with the attributes of the dependency.

    """

    attributes = {}

    for k in file.varattsget(depend_key):
        attributes[k] = file.varattsget(depend_key)[k]

        if isinstance(attributes[k], str) and attributes[k] in file.cdf_info()["zVariables"] \
                and not k == "DEPEND_0":
            try:
                attributes[k] = file.varget(attributes[k], starttime=tint[0], endtime=tint[1])
            except IndexError:
                attributes[k] = file.varget(attributes[k])
            # If atts is 2D get only first row
            if attributes[k].ndim == 2:
                try:
                    attributes[k] = attributes[k][0, :]
                except IndexError:
                    pass

    # Remove spaces in label
    try:
        attributes["LABLAXIS"] = attributes["LABLAXIS"].replace(" ", "_")

        if attributes["LABLAXIS"] == "Diffential_energy_channels":
            attributes["LABLAXIS"] = "Differential_energy_channels"

    except (KeyError, AttributeError):
        attributes["LABLAXIS"] = "comp"

    return attributes


def get_depend(file, cdf_name, tint, depend_num=1):
    """
    Read the `depend_num`th dependency data and its attributes associated to the variable named
    `cdf_name` in `file`.

    Parameters
    ----------
    file : cdflib.cdfread.CDF
        cdf file

    cdf_name : str
        Name of the target variable in the cdf file

    tint : list of str
        Time interval

    depend_num : int
        Index of the dependency

    Returns
    -------
    out : dict
        Hash table with the data of the dependency and its meta-data.

    """

    out = {}

    try:
        depend_key = file.varattsget(cdf_name)["DEPEND_{:d}".format(depend_num)]
    except KeyError:
        depend_key = file.varattsget(cdf_name)["REPRESENTATION_{:d}".format(depend_num)]

    if depend_key == "x,y,z":
        out["data"] = np.array(depend_key.split(","))

        out["atts"] = {"LABLAXIS": "comp"}
    else:
        try:
            out["data"] = file.varget(depend_key, starttime=tint[0], endtime=tint[1])
        except IndexError:
            out["data"] = file.varget(depend_key)

        out["data"] = file.varget(depend_key)

        if len(out["data"]) == 1:
            out["data"] = out["data"][0]

        if len(out["data"]) == 4 and all(out["data"] == ["x", "y", "z", "r"]):
            out["data"] = out["data"][:-1]

        elif out["data"].ndim == 2:
            if len(out["data"].flatten()) == 3:
                out["data"] = out["data"].flatten()
            else:
                try:
                    out["data"] = out["data"][0, :]
                except IndexError:
                    pass

        out["atts"] = get_depend_attributes(file, depend_key, tint)

    return out


def get_ts(file_path, cdf_name, tint):
    """
    Read field named cdf_name in file and convert to time series.

    Parameters
    ----------
    file_path : str
        Path of the cdf file.

    cdf_name : str
        Name of the target variable in the cdf file.

    tint : list of str
        Time interval.

    Returns
    -------
    out : xarray.DataArray
        Time series of the target variable in the selected time interval.

    """

    # Convert time interval to epochs
    tint = list(map(date_parser.parse, tint))
    tint = list(map(datetime_to_tt2000, tint))
    tint = list(map(cdfepoch.parse, tint))

    out_dict = {}
    time, depend_1, depend_2, depend_3 = [{}, {}, {}, {}]

    with CDF(file_path) as file:
        out_dict["atts"] = file.varattsget(cdf_name)

        assert "DEPEND_0" in out_dict["atts"] and "epoch" in out_dict["atts"]["DEPEND_0"].lower()

        time = get_epochs(file, cdf_name, tint)

        if "DEPEND_1" in out_dict["atts"] or "REPRESENTATION_1" in out_dict["atts"]:
            depend_1 = get_depend(file, cdf_name, tint, 1)

        elif "afg" in cdf_name or "dfg" in cdf_name:
            depend_1 = {"data": ["x", "y", "z"], "atts": {"LABLAXIS": "comp"}}

        if "DEPEND_2" in out_dict["atts"] or "REPRESENTATION_2" in out_dict["atts"]:
            depend_2 = get_depend(file, cdf_name, tint, 2)

            if depend_2["atts"]["LABLAXIS"] == depend_1["atts"]["LABLAXIS"]:
                depend_1["atts"]["LABLAXIS"] = "rcomp"
                depend_2["atts"]["LABLAXIS"] = "ccomp"

        if "DEPEND_3" in out_dict["atts"] or "REPRESENTATION_3" in out_dict["atts"]:
            assert out_dict["atts"]["REPRESENTATION_3"] != "x,y,z"

            depend_3 = get_depend(file, cdf_name, tint, 3)

            if depend_3["atts"]["LABLAXIS"] == depend_2["atts"]["LABLAXIS"]:
                depend_2["atts"]["LABLAXIS"] = "rcomp"
                depend_3["atts"]["LABLAXIS"] = "ccomp"

        if "sector_mask" in cdf_name:
            depend_1_key = file.varattsget(cdf_name.replace("sector_mask", "intensity"))["DEPEND_1"]

            depend_1["data"] = file.varget(depend_1_key)
            depend_1["atts"] = file.varattsget(depend_1_key)

            depend_1["atts"]["LABLAXIS"] = depend_1["atts"]["LABLAXIS"].replace(" ", "_")

        if "edp_dce_sensor" in cdf_name:
            depend_1["data"] = ["x", "y", "z"]
            depend_1["atts"] = {"LABLAXIS": "comp"}

        out_dict["data"] = file.varget(cdf_name, starttime=tint[0], endtime=tint[1])

        if out_dict["data"].ndim == 2 and out_dict["data"].shape[1] == 4:
            out_dict["data"] = out_dict["data"][:, :-1]

    if time and not depend_1 and not depend_2 and not depend_3:
        dims = ["time"]
        coords_data = [time["data"]]
        coords_atts = [time["atts"]]

    elif time and depend_1 and not depend_2 and not depend_3:
        dims = ["time", depend_1["atts"]["LABLAXIS"]]
        coords_data = [time["data"], depend_1["data"]]
        coords_atts = [time["atts"], depend_1["atts"]]

    elif time and depend_1 and depend_2 and not depend_3:
        if depend_1["atts"]["LABLAXIS"] == depend_2["atts"]["LABLAXIS"]:
            depend_1["atts"]["LABLAXIS"] = "rcomp"
            depend_2["atts"]["LABLAXIS"] = "ccomp"

        dims = ["time", depend_1["atts"]["LABLAXIS"], depend_2["atts"]["LABLAXIS"]]
        coords_data = [time["data"], depend_1["data"], depend_2["data"]]
        coords_atts = [time["atts"], depend_1["atts"], depend_2["atts"]]

    elif time and depend_1 and depend_2 and depend_3:
        if depend_2["atts"]["LABLAXIS"] == depend_3["atts"]["LABLAXIS"]:
            depend_2["atts"]["LABLAXIS"] = "rcomp"
            depend_3["atts"]["LABLAXIS"] = "ccomp"

        dims = ["time", depend_1["atts"]["LABLAXIS"], depend_2["atts"]["LABLAXIS"],
                depend_3["atts"]["LABLAXIS"]]
        coords_data = [time["data"], depend_1["data"], depend_2["data"], depend_3["data"]]
        coords_atts = [time["atts"], depend_1["atts"], depend_2["atts"], depend_3["atts"]]

    else:
        raise NotImplementedError

    out = xr.DataArray(out_dict["data"], coords=coords_data, dims=dims, attrs=out_dict["atts"])

    for dim, coord_atts in zip(dims, coords_atts):
        out[dim].attrs = coord_atts

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""avg_4sc.py
@author: Louis Richard
"""

from .resample import resample


def avg_4sc(b_list):
    """Computes the input quantity at the center of mass of the MMS
    tetrahedron.

    Parameters
    ----------
    b_list : list of xarray.DataArray
        List of the time series of the quantity for each spacecraft.

    Returns
    -------
    b_avg : xarray.DataArray
        Time series of the input quantity a the enter of mass of the
        MMS tetrahedron.

    Examples
    --------
    >>> from pyrfu.mms import get_data
    >>> from pyrfu.pyrf import avg_4sc

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft indices

    >>> b_mms = [get_data("B_gse_fgm_srvy_l2", tint, i) for i in range(1, 5)]
    >>> b_xyz = avg_4sc(b_mms)

    """

    b_list = [resample(b, b_list[0]) for b in b_list]

    b_avg = sum(b_list) / len(b_list)

    return b_avg



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

import numpy as np
import xarray as xr

from ..pyrf import ts_scalar, resample


def correct_edp_probe_timing(sc_pot):
    """Corrects for the channel delays not accounted for in the MMS EDP
    processing. As described in the MMS EDP data products guide.

    Parameters
    ----------
    sc_pot : xarray.DataArray
        Time series created from L2 sc_pot files, from the variable
        "mms#_edp_dcv_brst_l2" containing individual probe potentials.

    Returns
    -------
    v_corrected : xarray.DataArray
        Time series where the channel delay for each probe have been
        accounted and corrected for.

    Notes
    -----
    This function is only useful for Burst mode data. For the other
    telemetry modes (i.e. slow and fast) the channel delays are
    completely negligible and the interpolation and resampling applied
    here will have no effect other than possibly introduce numerical
    noise.

    """

    e_fact = [.1200, .1200, .0292]

    # Reconstruct E12, E34, E56 as computed in MMS processing
    time = sc_pot.time.data

    diff_sc_pot = []
    for i, fact in zip([0, 2, 4], e_fact):
        diff_sc_pot.append(ts_scalar(time, np.diff(sc_pot.data[:, i:i + 2])
                                     / fact))

    # Correct the time tags to create individual time series
    tau_vs = [np.timedelta64(0, "ns"), np.timedelta64(7629, "ns"),
              np.timedelta64(15259, "ns")]
    tau_es = [np.timedelta64(26703, "ns"), np.timedelta64(30518, "ns"),
              np.timedelta64(34332, "ns")]

    # Odds probes potential 1, 3, 5
    sc_pot_odds = []
    for tau, i in zip(tau_vs, [0, 2, 4]):
        sc_pot_odds.append(ts_scalar(time + tau, sc_pot.data[:, i]))

    # Electric field
    diff_sc_pot = []
    for e, tau in zip(diff_sc_pot, tau_es):
        diff_sc_pot.append(ts_scalar(e.time.data + tau, e.data))

    # Resample all data to time tags of V1 (i.e. timeOrig).
    sc_pot_odds = [resample(v, sc_pot) for v in sc_pot_odds]
    diff_sc_pot = [resample(e, sc_pot) for e in diff_sc_pot]

    # Recompute individual even probe potentials 2, 4, 6
    sc_pot_even = []
    for v, e, fact in zip(sc_pot_odds, diff_sc_pot, e_fact):
        sc_pot_even.append(v - e * fact)

    sc_pot_corrected = []
    for tup in zip(sc_pot_even, sc_pot_odds):
        for item in tup:
            sc_pot_corrected.append(item.data)

    # Create the new time series with the corrected values
    options = dict(coords=[time, np.arange(1, 7)], dims=["time", "probe"])
    sc_pot_corrected = xr.DataArray(sc_pot_corrected, **options)

    return sc_pot_corrected



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""ts_vec_xyz.py
@author: Louis Richard
"""

import xarray as xr


def ts_vec_xyz(time, data, attrs=None):
    """Create a time series containing a 1st order tensor.

    Parameters
    ----------
    time : ndarray
        Array of times.

    data : ndarray
        Data corresponding to the time list.

    attrs : dict
        Attributes of the data list.

    Returns
    -------
    out : xarray.DataArray
        1st order tensor time series.

    """

    assert data.ndim == 2 and data.shape[1] == 3
    assert len(time) == len(data), "Time and data must have the same length"

    if attrs is None:
        attrs = {}

    out = xr.DataArray(data, coords=[time[:], ["x", "y", "z"]],
                       dims=["time", "comp"], attrs=attrs)

    out.attrs["TENSOR_ORDER"] = 1

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""psd.py
@author: Louis Richard
"""

import warnings
import numpy as np
import xarray as xr

from scipy import signal


def psd(inp, n_fft=256, n_overlap=128, window="hamming", d_flag="constant",
        scaling="density"):
    """Estimate power spectral density using Welch's method.

    Welch's method [11]_ computes an estimate of the power spectral
    density by dividing the data into overlapping segments, computing a
    modified periodogram for each segment and averaging the
    periodograms.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of measurement values.

    n_fft : int, optional
        Length of the FFT used, if a zero padded FFT is desired.
        Default to 256.

    n_overlap : int, optional
        Number of points to overlap between segments. Default to 128.

    window : str, optional
        Desired window to use. It is passed to `get_window` to generate
        the window values, which are DFT-even by default.
        See "get_window" or a list of windows and required parameters.
        Default Hanning

    d_flag : str, optional
        Specifies how to detrend each segment. It is passed as the
        "type" argument to the"detrend" function.
        Default to "constant".

    scaling : str, optional
        Selects between computing the power spectral density
        ('density') where `Pxx` has units of V**2/Hz and computing the
        power spectrum ("spectrum") where "Pxx" has units of V**2, if
        `x` is measured in V and "fs" is measured in Hz.
        Default to 'density'

    Returns
    -------
    out : xarray.DataArray
        Power spectral density or power spectrum of inp.

    References
    ----------
    .. [11] P. Welch, "The use of the fast Fourier transform for the
            estimation of power spectra: A method based on time
            averaging over short, modified periodograms",
            IEEE Trans. Audio Electroacoust. vol. 15, pp. 70-73, 1967.

    """

    if inp.ndim == 2 and inp.shape[-1] == 3:
        inp = np.abs(inp)

    if n_overlap is None:
        n_persegs = 256
        n_overlap = n_persegs / 2
    else:
        n_persegs = 2 * n_overlap

    if n_fft < n_persegs:
        n_fft = n_persegs
        warnings.warn("nfft < n_persegs. set to n_persegs", UserWarning)

    f_samp = 1e9 / np.median(np.diff(inp.time.data)).astype(float)

    freqs, p_xx = signal.welch(inp.data, nfft=n_fft, fs=f_samp,
                               window=window, noverlap=n_overlap,
                               detrend=d_flag, nperseg=n_persegs,
                               scaling=scaling, return_onesided=True, axis=-1)

    out = xr.DataArray(p_xx, coords=[freqs], dims=["f"])

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""date_str.py
@author: Louis Richard
"""

from dateutil import parser


def date_str(tint, fmt=1):
    """Creates a string corresponding to time interval for output plot
    naming.

    Parameters
    ----------
    tint : list of str
        Time interval.

    fmt : int
        Format of the output :
            * 1 : "%Y%m%d_%H%M"
            * 2 : "%y%m%d%H%M%S"
            * 3 : "%Y%m%d_%H%M%S"_"%H%M%S"
            * 4 : "%Y%m%d_%H%M%S"_"%Y%m%d_%H%M%S"

    Returns
    -------
    out : str
        String corresponding to the time interval in the desired format.

    """

    start_time = parser.parse(tint[0])
    end_time = parser.parse(tint[1])

    if fmt == 1:
        out = start_time.strftime("%Y%m%d_%H%M")
    elif fmt == 2:
        out = start_time.strftime("%y%m%d%H%M%S")
    elif fmt == 3:
        out = "_".join([start_time.strftime("%Y%m%d_%H%M%S"),
                        end_time.strftime("%H%M%S")])
    elif fmt == 4:
        out = "_".join([start_time.strftime("%Y%m%d_%H%M%S"),
                        end_time.strftime("%Y%m%d_%H%M%S")])
    else:
        raise ValueError("Unknown format")

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.dates as mdates

from pandas.plotting import register_matplotlib_converters
register_matplotlib_converters()

plt.style.use("seaborn-whitegrid")
date_form = mdates.DateFormatter("%H:%M:%S")
sns.set_context("paper")

plt.rc("lines", linewidth=1)
colors = np.array([[0, 0, 0], [213, 94, 0], [0, 158, 115], [86, 180, 233]]) / 255
markers = ["s", "d", "o", "^"]


def mms_pl_config(r_mms):
    """Plot spacecraft configuaration with three 2d plots of the position in re and one 3d plot of
    the relative position of the spacecraft:

    Parameters
    ----------
    r_mms : list of xarray.DataArray
        Time series of the spacecraft position

    Returns
    -------
    fig : matplotlib.pyplot.figure
        to fill.

    axs : matplotlib.pyplot.subplotsaxes
        to fill.

    """

    r_earth = 6378.136
    r_xyz = np.vstack([np.mean(r_xyz, 0) for r_xyz in r_mms])
    r_xyz = np.mean(r_xyz, 0)
    delta_r = r_xyz - np.tile(r_xyz, (4, 1))

    fig = plt.figure(figsize=(9, 9))
    gs0 = fig.add_gridspec(3, 3, hspace=0.3, left=0.1, right=0.9, bottom=0.1, top=0.9)

    gs00 = gs0[0, :].subgridspec(1, 3, wspace=0.35)
    gs10 = gs0[1:, :].subgridspec(1, 1, wspace=0.35)

    axs0 = fig.add_subplot(gs00[0])
    axs1 = fig.add_subplot(gs00[1])
    axs2 = fig.add_subplot(gs00[2])
    axs3 = fig.add_subplot(gs10[0], projection='3d')

    earth = plt.Circle((0, 0), 1, color='k', clip_on=False)

    x_axis_labels = ["$X$ [$r_earth$]", "$Y$ [$r_earth$]", "$X$ [$r_earth$]"]
    y_axis_labels = ["$Z$ [$r_earth$]", "$Z$ [$r_earth$]", "$Y$ [$r_earth$]"]

    for axis, x_idx, y_idx, x_label, y_label in zip([axs0, axs1, axs2], [0, 1, 0], [2, 1, 1],
                                                    x_axis_labels, y_axis_labels):
        for i, marker in enumerate(markers):
            axis.scatter(r_xyz[i, x_idx] / r_earth, r_xyz[i, y_idx] / r_earth, marker=marker)

        axis.add_artist(earth)
        axis.set_xlim([20, -20])
        axis.set_ylim([-20, 20])
        axis.set_aspect("equal")
        axis.set_xlabel(x_label)
        axis.set_ylabel(y_label)

    axs3.view_init(elev=13, azim=-20)

    for i, marker in enumerate(markers):
        options = dict(s=50, marker=marker)
        axs3.scatter(delta_r[i, 0], delta_r[i, 1], delta_r[i, 2], **options)

        options = dict(color=colors[i], marker=marker, zdir='z', zs=-30)
        axs3.plot([delta_r[i, 0]] * 2, [delta_r[i, 1]] * 2, **options)
        axs3.plot([delta_r[i, 0]] * 2, [delta_r[i, 2]] * 2, **options)
        axs3.plot([delta_r[i, 1]] * 2, [delta_r[i, 2]] * 2, **options)

        options = dict(color="k", linestyle="--", linewidth=.5)
        axs3.plot([delta_r[i, 0]] * 2, [delta_r[i, 1]] * 2, [-30, delta_r[i, 2]], **options)
        axs3.plot([delta_r[i, 0]] * 2, [-30, delta_r[i, 1]], [delta_r[i, 2]] * 2, **options)
        axs3.plot([-30, delta_r[i, 0]], [delta_r[i, 1]] * 2, [delta_r[i, 2]] * 2, **options)

    for idx_0, idx_1 in zip([0, 1, 2, 0, 1, 2], [1, 2, 0, 3, 3, 3]):
        axs3.plot(delta_r[[idx_0, idx_1], 0], delta_r[[idx_0, idx_1], 1],
                  delta_r[[idx_0, idx_1], 2], 'k-')

    axs3.set_xlim([-30, 30])
    axs3.set_ylim([30, -30])
    axs3.set_zlim([-30, 30])
    axs3.set_xlabel("$\\Delta X$ [km]")
    axs3.set_ylabel("$\\Delta Y$ [km]")
    axs3.set_zlabel("$\\Delta Z$ [km]")

    axs3.legend(["MMS1", "MMS2", "MMS3", "MMS4"], frameon=False)

    axs = [axs0, axs1, axs2, axs3]

    return fig, axs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""datetime_to_tt2000.py
@author: Louis Richard
"""

import pandas as pd


def datetime_to_tt2000(time):
    """Transforms datetime to TT2000 string format.

    Parameters
    ----------
    time : datetime.datetime
        Time to convert to tt2000 string.

    Returns
    -------
    tt2000 : str
        Time in TT20000 iso_8601 format.

    """

    time_datetime = pd.Timestamp(time)

    # Convert to string
    tt2000 = "{}{:03d}".format(time_datetime.strftime('%Y-%m-%dT%H:%M:%S.%f'),
                              time_datetime.nanosecond)

    return tt2000
